/*
This file is a part of Lemick Basic (C) Copyright Alex Iliasov 2001.
Various support routines for translated programmes
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <time.h>

#define ERR_OPEN 1


struct FC_arr {
    long ssz, *dm;
    int dtp, dms;
    void *data;
};


char *strdup(const char *s);
void *mmalloc(long size);
void *m2malloc(long size);
void _go_gc();
void _2go_gc();
char *mstrdup(char *a);
char *FCes = "";
int mstrcmp(char *t1, char *t2);

/* dynamic array structure allocation and initialization */

void FC_arr_c(struct FC_arr *_fa_x, int dtp, long ssz, int dms, ...)
{
    int i = 0;
    long ts = 1;
    va_list ap;
    va_start(ap, dms);

    _fa_x->ssz = ssz;						     /* size of element (bytes) */
    _fa_x->dtp = dtp;						     /* code of datatype */
    _fa_x->dms = dms;						     /* number of dims */
    _fa_x->dm = (long *) m2malloc(sizeof(long) * dms);		     /* array of dimens. */
    while (i < dms) {
	_fa_x->dm[i] = va_arg(ap, long);
	ts *= _fa_x->dm[i];
	i++;
    }
/* ts*=10;*/

    _fa_x->data = m2malloc(ssz * ts);				     /* array data */
    va_end(ap);
}

void FC_arr_init1(void **_fa_x, void (*init) (void *), long ts, long ssz)
{
    int i;
    for (i = 0; i < ts; i++) {
	_fa_x[i] = m2malloc(ssz);
	(*init) ((void *) _fa_x[i]);
    }
}

void FC_arr_init2(struct FC_arr *_fa_x, void (*init) (void *))
{
    long ts = 1;
    int i;
    for (i = 0; i < _fa_x->dms; i++)
	ts *= ts * _fa_x->dm[i];
    for (i = 0; i < ts; i++)
	(*init) ((char *) _fa_x->data + i * _fa_x->ssz);
}



char *Fas(char *s1, char *s2)
{
    char *t;
    if ((s1 == NULL || s1 == FCes) && (s2 == NULL || s2 == FCes))
	return FCes;
    if (s1 == NULL || s1 == FCes)
	return mstrdup(s2);
    if (s2 == NULL || s2 == FCes)
	return mstrdup(s1);
    t = (char *) mmalloc(strlen(s1) + strlen(s2) + 1);
    strcpy(t, s1);
    strcat(t, s2);
    return t;
}



char *Ff2s(long double vl)
{
    char *t;
    t = (char *) mmalloc(50);
    sprintf(t, "%Lg", vl);
    return t;
}

char *Fi2s(long vl)
{
    char *t;
    t = (char *) mmalloc(50);
    sprintf(t, "%ld", vl);
    return t;
}

long double Fs2f(char *s)
{
    if (s == FCes || s == NULL)
	return 0;
    return atof(s);
}

void Fss(char **trg, char *vl)
{
    if (vl == NULL || vl == FCes) {
	*trg = 0;
	return;
    }
    *trg = (char *) mmalloc(strlen(vl) + 1);
    strcpy(*trg, vl);
}


char *Fleft(const char *s, long ln)
{
    char *t;
    if (s == FCes || s == NULL || ln <= 0)
	return FCes;
    t = (char *) mmalloc(ln + 1);
    if (ln > strlen(s))
	ln = strlen(s);
    strncpy(t, s, ln);
    t[ln] = '\0';
    return t;
}

char *Fright(char *s, long ln)
{
    char *t;
    char *tt;
    if (s == FCes || s == NULL || ln <= 0)
	return FCes;
    if (ln > strlen(s))
	ln = strlen(s);
    t = s + strlen(s) - ln;
    tt = (char *) mmalloc(ln + 1);
    strncpy(tt, t, ln);
    tt[ln] = '\0';
    return tt;
}



#if defined(__cplusplus)
/*thou there no inline functions in ansi C, gcc (as it's also c++ aware)
has such a feature.*/
inline char *_sw(char *s)
{
    if (!s)
	return FCes;
    else
	return s;
}

#else
char *_sw(char *s)
{
    if (!s)
	return FCes;
    else
	return s;
}
#endif

void _msw(char **s)
{
    char *t;
    if (*s) {
	t = mmalloc(strlen(*s) + 1);
	strcpy(t, *s);
	free(*s);
	*s = t;
    }
}

/*seems like very ineffecient implementation*/

void _rs(FILE * f, char **s)
{
    char *t, *ot, *u = NULL;
    int c;
    long k;
    t = calloc(2048, 1);
    ot = t;
    while (!feof(f)) {
	c = fgetc(f);
	if (c == EOF || c == ',')
	    break;
	*t = c;
	t++;
    }
    k = strlen(ot);
    u = mmalloc(k + 1);
    strcpy(u, ot);
    *s = u;
}

void _rl(FILE * f, char **s)
{
    char *t, *u = NULL;
    long k;
    t = malloc(2048);
    if (!fgets(t, 2048, f)) {
	*s = 0;
	return;
    }
    k = strlen(t);
    u = mmalloc(k);
    if (t[k - 1] == '\n')
	k--;
    strncpy(u, t, k);
    free(t);
    *s = u;
}


char *Fmmid(const char *s, int st, int ln)
{
    char *t;
    if (s == FCes || s == NULL)
	return FCes;
    if ((ln == -1) || ((st + ln) > strlen(s)))
	ln = strlen(s) - st;
    if ((st > strlen(s)) || (st < 0) || (ln < 0) || (strlen(s) == 0))
	return FCes;
    t = (char *) mmalloc(ln + 1);
    strncpy(t, s + st, ln);
    t[ln] = '\0';
    return t;
}

char *Fltrim(char *s)
{
    char *u;
    if (s == FCes || s == NULL)
	return FCes;
    while (*s == ' ')
	s++;
    u = (char *) mmalloc(strlen(s) + 1);
    strcpy(u, s);
    return u;
}

char *Frtrim(char *s)
{
    int t;
    char *u;
    if (s == FCes || s == NULL)
	return FCes;
    t = strlen(s) - 1;
    while (*(s + t) == ' ')
	t--;
    u = (char *) mmalloc(t + 2);
    strncpy(u, s, t + 1);
    u[t + 1] = '\0';
    return u;
}

long Finstr(char *t, char *t2, int start)
{
    char *t3;
    if ((t == FCes || t == NULL) || (t2 == FCes || t2 == NULL))
	return -1;
    if (start > strlen(t))
	return -1;
    t3 = strstr(t + start, t2);
    if (t3 != NULL)
	return (long) (t3 - t);
    else
	return -1;
}

int mstrcmp(char *t1, char *t2)
{
    if ((t1 == FCes || t1 == NULL) || (t2 == FCes || t2 == NULL))
	return 0;
    if (t1 == FCes || t1 == NULL)
	return -1;
    if (t2 == FCes || t2 == NULL)
	return 1;
    return strcmp(t1, t2);
}

int mstrlen(char *t1)
{
    if (t1 == FCes || t1 == NULL)
	return 0;
    return strlen(t1);
}


/*very simplistic gc */

typedef struct {
    void *p;
    void *next;
} gct;

gct first = { NULL, NULL };
gct *now = &first;
gct first2 = { NULL, NULL };
gct *now2 = &first2;

void _2gc(void *p)
{
/*  static int kk=0; */
    now->next = malloc(sizeof(gct));
    now = (gct *) now->next;
    now->p = p;
    now->next = NULL;
/*  printf("[G:%d]",kk++); */
}

void _go_gc()
{
    gct *cp, *ocp;
    cp = (gct *) first.next;
    while (cp) {
	free(cp->p);
	ocp = cp;
	cp = (gct *) cp->next;
	free(ocp);
/*    printf("(g)"); */
    }
    first.next = NULL;
    now = &first;
}

void _22gc(void *p)
{
    now2->next = malloc(sizeof(gct));
    now2 = (gct *) now2->next;
    now2->p = p;
    now2->next = NULL;
}

void _2go_gc()
{
    gct *cp, *ocp;
    cp = (gct *) first2.next;
    while (cp) {
	free(cp->p);
	ocp = cp;
	cp = (gct *) cp->next;
	free(ocp);
    }
    first2.next = NULL;
    now2 = &first2;
}


void *mmalloc(long size)
{
    void *p;
    if ((p = calloc(size, 1)) == NULL) {
	fprintf(stderr, "Failed to allocate memory\n");
	exit(0);
    }
    _2gc(p);
    return p;
}

char *mstrdup(char *a)
{
    char *t;
    if (a == FCes || a == NULL)
	return FCes;
    t = mmalloc(strlen(a) + 1);
    strcpy(t, a);
    return t;
}


void *m2malloc(long size)
{
    void *p;
    if ((p = calloc(size, 1)) == NULL) {
	fprintf(stderr, "Failed to allocate memory\n");
	exit(0);
    }
    _22gc(p);
    return p;
}

typedef struct fh_t_ {
    FILE *s;
    int i;
    int code;
} fh_t;


fh_t *fhs = NULL;
int fs = 0;
int fc = -1;

void newfh(FILE * f, int i, int code)
{
    fc++;
    if (fc >= fs) {
	fs += 32;
	fhs = realloc(fhs, fs);
    }
    fhs[fc].s = f;
    fhs[fc].i = i;
    fhs[fc].code = code;
}

FILE *getfh_s(int code)
{
    int i;
    for (i = 0; i <= fc; i++)
	if (fhs[i].code == code)
	    return fhs[i].s;
    fprintf(stderr, "Bad file number\n");
    exit(0);
}

int getfh_i(int code)
{
    int i;
    for (i = 0; i <= fc; i++)
	if (fhs[i].code == code)
	    return fhs[i].i;
    fprintf(stderr, "Bad file number\n");
    exit(0);
}




char *Fdate()
{
    char *d;
    time_t x;
    struct tm *y;
    d = (char *) mmalloc(11);
    x = time(&x);
    y = localtime(&x);
    sprintf(d, "%2d.%2d.%4d", y->tm_mday, y->tm_mon, y->tm_year + 1900);
    return d;
}

char *Ftime()
{
    char *d;
    time_t x;
    struct tm *y;
    d = (char *) mmalloc(9);
    x = time(&x);
    y = localtime(&x);
    sprintf(d, "%2d:%2d:%2d", y->tm_hour, y->tm_min, y->tm_sec);
    return d;
}

long Flof(FILE * fp)
{
    long ts, res;
    ts = ftell(fp);
    fseek(fp, 0, SEEK_END);
    res = ftell(fp);
    fseek(fp, ts, SEEK_SET);
    return res;
}

char *Fread(long len, FILE * fp)
{
    char *t;
    t = (char *) malloc(len + 1);
    fread((void *) t, 1, len, fp);
    return t;
}

char *Flread(FILE * fp, long len)
{
    char *t;
    t = (char *) malloc(len + 1);

    if (fgets(t, len, fp) == NULL) {
	free(t);
	return NULL;
    };
    if (*(t + strlen(t) - 1) == '\n')
	*(t + strlen(t) - 1) = '\0';
    return t;
}
