/*
This file is a part of Lemick Basic (C) Copyright Alex Iliasov 2001.
Data holding structures, like stacks and hashes.
Idea for parallel stack `struct _stck_t' taken from mine sfos programme
*/

#include <stdlib.h>
#include "glb.h"
#include "sym.h"

extern char *strdup(char *);
extern int strcasecmp(char *, char *);

struct _stck_t {
    char *buf, *cbuf, *ebuf, *sbuf, *elast, *lke;
    etp *buf2;
    int buf2p, ebuf2, lkt;
    char *name;
};
typedef struct _stck_t stck;

struct _simtck_t {
    int *buf;
    int bufp, ebuf;
    char *name;
};
typedef struct _simtck_t simtck;

struct _hash_t {
    char **sval;
    long *tval;
    long pos, maxpos;
    char *name;
};
typedef struct _hash_t hash;



hash *new_buf_hash(long sz1, char *name)
{
    hash *t;
    t = (hash *) malloc(sizeof(hash));
    t->sval = (char **) malloc(sz1 * sizeof(char *));
    t->tval = (long *) malloc(sz1 * sizeof(long));
    t->name = strdup(name);
    t->pos = -1;
    t->maxpos = sz1;
    return t;
}

void addh(hash * t, char *n, long v)
{
//  printf("in addh:%s,%d\n",n,v);
    if (t->pos + 1 == t->maxpos) {
	printf("<%s> buffer overflow\n", t->name);
	exit(0);
    }
    t->pos++;
    t->sval[(t->pos)] = strdup(n);
    t->tval[(t->pos)] = v;
}

void addhr(hash * t, char *n, long v)
{
    long i = t->pos;
//  printf("in addhr:%s,%d\n",n,v);
    if (i > -1)
	do {
	    if (t->tval[i] == v)
		break;
	    i--;
	} while (i > -1);
    if (i == -1)
	i = ++t->pos;
    else
	free(t->sval[(t->pos)]);
    if (t->pos == t->maxpos) {
	printf("<%s> buffer overflow\n", t->name);
	exit(0);
    }
    t->sval[i] = strdup(n);
    t->tval[i] = v;
}


void freeh(hash * t)
{

    if ((t->pos) == -1)
	return;
    do {
	free(t->sval[(t->pos)]);
	t->pos--;
    } while ((t->pos) > -1);
}

void deleteh(hash * t)
{
    freeh(t);
    free(t);
}

long hash_by_name(hash * t, char *name)
{
    long i = (t->pos);
    do {
	if (!strcmp(t->sval[i], name))
	    return t->tval[i];
	i--;
    } while (i > -1);
    return -1;
}

long hash_by_iname(hash * t, char *name)
{
    long i = t->pos;
    do {
	if (!strcasecmp(t->sval[i], name))
	    return t->tval[i];
	i--;
    } while (i > -1);
    return -1;
}

char *hash_by_val(hash * t, long val)
{
    long i = (t->pos);
    do {
	if (t->tval[i] == val)
	    return t->sval[i];
	i--;
    } while (i > -1);
    return "";
}


stck *new_buf_stack(long sz1, long sz2, char *name)
{
    stck *t;
    t = (stck *) malloc(sizeof(stck));
    t->buf = (char *) malloc(sz1 * sizeof(char));
    t->buf2 = (etp *) malloc(sz2 * sizeof(etp));
    t->name = strdup(name);
    t->buf[0] = '\0';
    t->buf[1] = '\0';
    t->cbuf = t->buf + 2;
    t->ebuf = t->buf + sz1 - 3;
    t->ebuf2 = sz2 - 1;
    t->buf2p = -1;
    t->elast = t->buf;
    t->lkt = -1;
    return t;
}

simtck *new_buf_sim_stack(long sz1, char *name)
{
    simtck *t;
    t = (simtck *) malloc(sizeof(simtck));
    t->buf = (int *) malloc(sz1 * sizeof(int));
    t->name = strdup(name);
    t->ebuf = sz1 - 1;
    t->bufp = -1;
    return t;
}


void flush_buf(stck * t)
{
    t->cbuf = t->buf + 2;
    t->buf2p = -1;
    t->elast = t->buf;
    t->lke = t->cbuf;
    t->lkt = -1;
}

void free_buf(stck * t)
{
    free(t->buf);
    free(t->buf2);
    free(t->name);
    free(t);
}

void flush_sbuf(simtck * t)
{
    t->bufp = -1;
}

void free_sbuf(simtck * t)
{
    free(t->buf);
    free(t->name);
    free(t);
}


void push(simtck * b, int v)
{
    b->bufp++;
    if (b->bufp > b->ebuf) {
	printf("Fault: %s buffer overflow\n", b->name);
	exit(0);
    }
    b->buf[b->bufp] = v;
}

int pop(simtck * b)
{
    if (b->bufp < 0)
	printf("%s: Buffer underflow\n", b->name);
    b->bufp--;
    return b->buf[b->bufp + 1];
}


void pushb(stck * b, char *s, etp tt)
{
    unsigned int t = strlen(s);
//  printf("push");
//  printf("(%s): push \0%s/(%d,%d)\n",s,tt.cl,tt.tp);
    if (b->cbuf + t >= b->ebuf) {
	printf("Fault: %s buffer(S) overflow\n", b->name);
	exit(0);
    }
    b->buf2p++;
    b->lkt = b->buf2p;
    if (b->buf2p > b->ebuf2) {
	printf("Fault: %s buffer(T) overflow\n", b->name);
	exit(0);
    }
    b->buf2[b->buf2p] = tt;
//  memmove(cbuf,s);
//  *(cbuf+t)='\0';
    strcpy(b->cbuf, s);
//  printf("Pushed <%s> : %s\n",s,buf+1);
    b->elast = b->cbuf;
    b->cbuf += t + 1;
    b->lke = b->cbuf;
}

void pushbe(stck * b, char *s)
{
    unsigned int t = strlen(s);
//  printf("push E\n");
    if (b->cbuf + t >= b->ebuf) {
	printf("Fault: compile buffer overflow\n");
	exit(0);
    }
    strcpy(b->cbuf, s);
    b->elast = b->cbuf;
    b->cbuf += t + 1;
    b->lke = b->cbuf;
}


void pushbt(stck * b, etp tt)
{
//  printf("push T\n");
    b->buf2p++;
    b->lkt = b->buf2p;
    if (b->buf2p > b->ebuf2) {
	printf("Fault: type buffer overflow\n");
	exit(0);
    }
    b->buf2[b->buf2p] = tt;
}


////1 2 3 4 5 6
/// 0|2|0|3|0|.|.|.|.
///           ^
///       ^
/// 1) 6; t=4; t=3; cbuf=4; 4;
char *popb(stck * b, etp * tt)
{
    char *t;
//  printf("(%s): popping\n");
    if ((b->cbuf - b->buf) <= 2 || b->buf2p == -1)
	return "";
    for (t = b->cbuf - 2; *t; t--);
    b->cbuf = t + 1;
    b->lke = b->cbuf;
    *tt = b->buf2[b->buf2p];
    b->buf2p--;
    b->lkt = b->buf2p;
//  printf("Poped <%s>:\n",t+1);
    return t + 1;
}


char *popbe(stck * b)
{
    char *t, *ot;
//  printf("(%s) popping E\n",b->name);
    if ((b->cbuf - b->buf) <= 2)
	return "";
    ot = b->cbuf;
    for (t = b->cbuf - 2; *t; t--);
    b->cbuf = t + 1;
    b->lke = b->cbuf;
    return t + 1;
}

etp popbt(stck * b)
{
//  printf("(%s) popping T\n",b->name);
    if (b->buf2p == -1)
	printf("Buffer underflow\n");
    b->buf2p--;
    b->lkt = b->buf2p;
    return b->buf2[b->buf2p + 1];
}

void begin_scan(stck * b)
{
    b->sbuf = b->buf + 2;
}

char *scan(stck * b)
{
    char *t;
    if (b->sbuf >= b->cbuf)
	return "";
    t = b->sbuf;
    b->sbuf += strlen(b->sbuf) + 1;
    return t;
}

char *lscan(stck * b, etp * tt)
{
    char *t;
    if (b->lke >= b->cbuf)
	return "";
    t = b->lke;
    b->lke += strlen(b->lke) + 1;
    *tt = b->buf2[b->lkt];
    b->lkt++;
    return t;
}

etp lookback(stck * b)
{
    char *t;
    if ((b->lke - b->buf) <= 2 || b->lkt == -1) {
	printf("buffer underflow!\n");
	exit(0);
    }
    for (t = b->lke - 2; *t; t--);
    b->lke = t + 1;
    b->lkt--;
    return b->buf2[b->lkt + 1];
}

void setl(stck * b)
{
    b->buf2p = b->lkt;
    b->cbuf = b->lke;
}

/*
void stack_show(stck* b)
 {
  char* t=b->buf+2;
  int i;
  printf("\nstack[%d;%d]:<<",b->buf2p,b->cbuf-b->buf);
  for(i=0;i<=b->buf2p;i++)
   {
    printf("%s(%d)|",t,b->buf2[i].tp); 
    t+=strlen(t)+1;
   }
   printf(">>\n");
 }
*/

/*
void pushbf(const char* fs, ...)
 {
  char* oc,*st;
  va_list ap;
  va_start(ap,fs);
  *oc=cbuf;
  *cbuf='\0';
  while (*fs) {
   switch(*fs) {
    case '%':{
       
       switch(*fs) {
        case 's':{strcat(oc,va_arg(ap,char *));oc=oc+strlen(oc);break;}
        case 'd':{sprintf(oc,"%d",va_arg(ap,int));break;}
        case 'l':{sprintf(oc,"%ld",va_arg(ap,long int));break;}
        case 'c':{sprintf(oc,"%c",va_arg(ap,char));break;}
       }
    }
   }
   fs++;
  }
  va_end(ap);
  return mplx_list;
 }

*/

/*

                  C
  		2*3+9
        C=L       +       C=R
        2*3                9
   C=L   *    C=R
    2          3	       
*/
