/*
This file is a part of Lemick Basic (C) Copyright Alex Iliasov 2001.
implementation of SELECT CASE Basic's structure (in style of VB)
*/

#include "glb.h"
#include "sym.h"
#include "cbufs.h"
#include "logis.h"

#define is_iter_type(x) (x==tINT || x==tDBL)


extern stck *stk0;
extern char *strdup(char *);

int sc = -1, ss = 0;
etp se[25];
char *sn[25];

void scase_start()
{
    etp tl;
    char *l;
    l = popb(stk0, &tl);
    if (tl.cl == cCONST)
	msgrep(mWRN, "constant expression in SELECT CASE");
    if (!is_iter_type(tl.tp) && tl.tp != tSTR)
	msgrep(mERR, "bad expression type in SELECT CASE");
    fprintf(out, "do {\n");
    sc++;
    se[sc] = tl;
    sn[sc] = strdup(l);
}

void scase_end()
{
    fprintf(out, "} while(0);\n");
    free(sn[sc]);
    sc--;
}

void scase_case_end()
{
    fprintf(out, "break;}\n");
}

void scase_case_plain()
{
    etp tl;
    char *l;
    l = popb(stk0, &tl);
    if (tl.cl > cSCALAR)
	msgrep(mERR, "Invalid operand class for CASE");
    switch (se[sc].tp) {
    case tINT:
    case tDBL:
	if (tl.tp == tSTR)
	    msgrep(mERR, "Invalid operand type for CASE");
	fprintf(out, "if (%s==%s) {\n", sn[sc], l);
	break;
    case tSTR:
	if (tl.tp != tSTR)
	    msgrep(mERR, "Invalid operand type for CASE");
	fprintf(out, "if (!mstrcmp(%s,%s)) {\n", sn[sc], l);
	break;
    }
}

void scase_case_range()
{
    fprintf(out, "if (");
    ss = 0;
}

void scase_case_range_end()
{
    fprintf(out, ") {\n");
}


void scase_case_range_single()
{
    etp tl;
    char *l;
    l = popb(stk0, &tl);
    if (tl.cl > cSCALAR)
	msgrep(mERR, "Invalid operand class for CASE");
    if (ss > 0)
	fprintf(out, " || ");
    ss++;
    switch (se[sc].tp) {
    case tINT:
    case tDBL:
	if (tl.tp == tSTR)
	    msgrep(mERR, "Invalid operand type for CASE");
	fprintf(out, "(%s==%s)", sn[sc], l);
	break;
    case tSTR:
	if (tl.tp != tSTR)
	    msgrep(mERR, "Invalid operand type for CASE");
	fprintf(out, "(!mstrcmp(%s,%s))", sn[sc], l);
	break;
    }
}

void scase_case_range_double()
{
    etp tl, tr;
    char *l, *r;
    l = strdup(popb(stk0, &tl));
    r = strdup(popb(stk0, &tr));
    if (tl.cl > cSCALAR || tr.cl > cSCALAR)
	msgrep(mERR, "Invalid operand class for CASE");
    if (ss > 0)
	fprintf(out, " || ");
    ss++;
    switch (se[sc].tp) {
    case tINT:
    case tDBL:
	if (tl.tp == tSTR || tr.tp == tSTR)
	    msgrep(mERR, "Invalid operand type for CASE");
	fprintf(out, "(%s<=%s && %s>=%s)", sn[sc], l, sn[sc], r);
	break;
    case tSTR:
	if (tl.tp != tSTR)
	    msgrep(mERR, "Invalid operand type for CASE");
	fprintf(out, "(mstrcmp(%s,%s)<=0 && mstrcmp(%s,%s)>=0)", sn[sc], l,
		sn[sc], r);
	break;
    }
    free(l);
    free(r);
}


void scase_case_binops(int bop)
{
    etp tl;
    char *l, *op = "==";
    l = popb(stk0, &tl);
    if (tl.cl > cSCALAR)
	msgrep(mERR, "Invalid operand class for CASE");
    switch (bop) {
    case 0:
	op = "==";
	break;
    case 1:
	op = ">";
	break;
    case 2:
	op = "<";
	break;
    case 3:
	op = "!=";
	break;
    case 4:
	op = "<=";
	break;
    case 5:
	op = ">=";
    }
    switch (se[sc].tp) {
    case tINT:
    case tDBL:
	if (tl.tp == tSTR)
	    msgrep(mERR, "Invalid operand type for CASE");
	fprintf(out, "if (%s%s%s) {\n", sn[sc], op, l);
	break;
    case tSTR:
	if (tl.tp != tSTR)
	    msgrep(mERR, "Invalid operand type for CASE");
	fprintf(out, "if (mstrcmp(%s,%s)%s0) {\n", sn[sc], l, op);
	break;
    }
}
