/*
This file is a part of Lemick Basic (C) Copyright Alex Iliasov 2001.
Main part of compiler. Various code for working with symbols table 
*/

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <math.h>
#include "sym_arr.h"
#include "glb.h"

#define getch()  fgetc(stdin)

#define DIMS(x)   arp[x].dims
#define TP(x)     arp[x].tp
#define FDIMS(x)  arp[symtab[x].ttp.tp].dims
#define FTP(x)    arp[symtab[x].ttp.tp].tp
#define FF(x)     ((arr_d*) symtab[x].adata)->flg
#define SNTH(x,y) ((arr_d*) symtab[x].adata)->d.dim_v[y]	     // Y-th dimension of X-th array (static)
#define DNTH(x,y) ((arr_d*) symtab[x].adata)->d.dim_e[y]	     // Y-th dimension of X-th array (dynamic)

#define FUNCF(x,y) (((fun_t*) symtab[x].adata)->flg & y)


extern int exp_count, is_array;

struct etp_ {
    int tp;
    int cl;
};
typedef struct etp_ etp;

struct ir_etp_ {
    etp t;
    void *adata;
};
typedef struct ir_etp_ ir_etp;

struct symt_ {
    char *name;
    int flg;
    int scope;
    etp ttp;
    void *adata;
};
typedef struct symt_ symt;

struct usert_ {
    long agr;
    int flg;
    long symp;
    int extc;
    int *ext;
};
typedef struct usert_ usert;



enum tscp { sPROJ = 0, sFILE, sGLOB, sMAIN, sSUB, sOTHER };

enum name_flags {
    nf_BYVAL = 1 << 0,
    nf_BYREF = 1 << 1,
    nf_OPTIONAL = 1 << 3,
    nf_EXTYPE = 1 << 4,
    nf_ARTYPE = 1 << 5,
    nf_PROTECTED = 1 << 6,
    nf_EXTERN = 1 << 7
};

enum type_flags {
    tp_WINIT = 1 << 0,
    tp_WKILL = 1 << 1
};



int TP_STRUCTS = cARRAY;
const etp s_NAME_SC = { cSCALAR, cSPEC };
const etp s_NAME_AR = { cARRAY, cSPEC };
const etp s_SYMB = { stANY, cSPEC };

const etp CT_STR = { tSTR, cCONST };
const etp CT_INT = { tINT, cCONST };
const etp CT_DBL = { tDBL, cCONST };
const etp SC_INT = { tINT, cSCALAR };
const etp SC_DBL = { tDBL, cSCALAR };


#define TBASE tDBL
long symf = 0, last_symp = 0;
int namef = 0;
char *_cmpt;
char spref[64];
int last_class = cARRAY;
int last_type = tMAX - 1;
#include "sym_fun.h"
#include "funreg.h"
#include "cbufs.h"
#include "igo.h"
extern stck *stk0, *stk1;

int cscope = 0;
symt *symtab;
usert *usrtab;
hash *typtab;
long usrp = -1;
long symp = -1;
int dims = -1;
#define SYMTAB_SIZE 200
char *lcase(char *s);
char cmpl_pref(int cl);

void print_cmpl_type(FILE * wt, int t);
void print_init_val(FILE * wt, long p);
char *sprint_cmpl_type(char *wt, int t);
void symtab_dec(long p);
etp metp(int t, int c);
long arr_static_size(long j);
void use_constructor(char *name, int tp, int pc);
long look_sub(char *name, int scope, int pc);
long sym_sizes(int t);


FILE *lfh;
extern int _nc;

void create_symtab()
{
    symtab = (symt *) malloc(SYMTAB_SIZE * sizeof(symt));
    usrtab = (usert *) malloc(SYMTAB_SIZE * sizeof(usert));
    typtab = new_buf_hash(SYMTAB_SIZE, "types table");
    addh(typtab, "int", tINT);
    addh(typtab, "word", tWORD);
    addh(typtab, "long", tLONG);
    addh(typtab, "dword", tDWORD);
    addh(typtab, "byte", tBYTE);
    addh(typtab, "real", tREAL);
    addh(typtab, "double", tDBL);
    addh(typtab, "string", tSTR);
    *spref = 0;
    init_ar();
    init_fn();
}

void free_symtab()
{
    free(symtab);
    free(usrtab);
    deleteh(typtab);
    free_ar();
}

inline int symtab_gate(const char *name, int scope, long i)
{
    if (!strcasecmp(symtab[i].name, name)) {
	if (symtab[i].flg == nf_PROTECTED) {
	    if (!(symf & sf_INSIDE_TYPE)) {
		if (scope == symtab[i].scope)
		    msgrep(mERR, "Can't access PRIVATE class member - %s",
			   name);
	    } else {
		if (symtab[i].scope == last_type)
		    return 1;
	    }
	    if (scope >= tMAX && usrtab[scope - tMAX].extc) {
		for (i = 0; i < usrtab[scope - tMAX].extc; i++)
		    if (symtab[i].scope ==
			usrtab[symtab[i].scope - tMAX].ext[i])
			return 1;
	    }
	    return 0;
	} else {
	    if (symf & sf_INSIDE_TYPE) {
		if (symtab[i].scope == last_type)
		    return 1;
	    }
	    if (scope == symtab[i].scope || symtab[i].scope == 1)
		return 1;
	    if (scope >= tMAX && usrtab[scope - tMAX].extc) {
		for (i = 0; i < usrtab[scope - tMAX].extc; i++)
		    if (symtab[i].scope == usrtab[scope - tMAX].ext[i])
			return 1;
	    }

	}
    }
    return 0;
}


long symtab_look(const char *name, int scope)
{
    long i;
    for (i = 0; i <= symp; i++)
	if (symtab_gate(name, scope, i))
	    return i + 1;
    return 0;
}


int symtab_type(long i)
{
    return symtab[i].ttp.tp;
}

int symtab_scope(long i)
{
    return symtab[i].scope;
}

int symtab_class(long i)
{
    return symtab[i].ttp.cl;
}

void usrtab_sf(int i, int f)
{
    usrtab[i].flg |= f;
}


char *symtab_name(long i)
{
    return symtab[i].name;
}

void *symtab_adata(long i)
{
    return symtab[i].adata;
}

void symtab_set_adata(long i, void *e)
{
    symtab[i].adata = e;
}

void symtab_set_flg(long i, int e)
{
    symtab[i].flg |= e;
}

void symtab_drop_flg(long i, int e)
{
    symtab[i].flg &= ~e;
}

int symtab_test_flg(long i, int f)
{
    return symtab[i].flg & f;
}

int symtab_flg(long i)
{
    return symtab[i].flg;
}

void sym_set_type(char *nm, int tp)
{
    long i;
    i = symtab_look(nm, cscope);
    if (!i)
	msgrep(mHLT, "Internal compiler error:sym_set_type %s/%d",
	       __FILE__, __LINE__);
    symtab[i - 1].ttp.tp = tp;
}


void write_subtypes_init(FILE * f, int tp)
{
    int i, k;
    for (i = 0; i <= symp; i++) {
	if ((symtab[i].scope == tp) && (symtab[i].ttp.tp >= tMAX)
	    && (symtab[i].ttp.cl == cSCALAR || symtab[i].ttp.cl == cARRAY)) {
	    if (usrtab[symtab[i].ttp.tp - tMAX].flg & tp_WINIT) {
		k = look_sub("init", symtab[i].ttp.tp, 0);
		if (k == -1)
		    continue;
		fprintf(f,
			"\n%s_fs_%s=(_fT_%s*) m2malloc(sizeof(_fT_%s));",
			spref, symtab[i].name, hash_by_val(typtab,
							   symtab[i].ttp.
							   tp),
			hash_by_val(typtab, symtab[i].ttp.tp));
		fprintf(f, "\n_%s_ff_%d_init(%s_fs_%s);",
			hash_by_val(typtab, symtab[i].ttp.tp), k, spref,
			symtab[i].name);
	    }
	}
    }
}



void delete_scope(int s)
{
    int mys;
    mys = symtab[symp].scope;
    while (mys == s) {
	free(symtab[symp].name);
	symp--;
	mys = symtab[symp].scope;
    }
}

void tumble()
{
    char *s1, *s2, *ss;
    etp t, tt;
    int i;
    ss = hash_by_val(ctx, ctx_PART_NAME);
    if (!strlen(ss))
	return;							     //then it was a method call, not array 
    s2 = strdup(popb(stk0, &t));
    s1 = strdup(popb(stk0, &tt));
    if (!(i = symtab_look(ss, cscope)))
	msgrep(mERR, "Undeclared variable - %s", ss);;
    last_symp = i - 1;
    if (tt.tp < tMAX)
	sprintf(stk0->cbuf, "%s.%s", s1, s2);
    else
	sprintf(stk0->cbuf, "%s->%s", s1, s2);
    stk0->cbuf += strlen(stk0->cbuf) + 1;
    free(s2);
    free(s1);
    pushbt(stk0, t);
    symf &= ~sf_COMPLX_NAME;
}


long sym_get_type(char *n)
{
    long t;
    t = hash_by_iname(typtab, n);
    if (t == -1)
	msgrep(mERR, "No such a type - %s", n);
    return t;
}

void sym_add_ext(char *n)
{
    int t;
    t = hash_by_iname(typtab, n);
    if (t == -1)
	msgrep(mERR,
	       "(in EXTENDS) Not a valid class name or class is undeclared - `%s'",
	       n);
    usrtab[usrp].ext =
	(int *) realloc(usrtab[usrp].ext,
			(usrtab[usrp].extc + 1) * sizeof(int));
    usrtab[usrp].ext[usrtab[usrp].extc] = t;
    usrtab[usrp].extc++;
}

void sym_enter_type(char *n)
{
    lfh = dec;
    dec = tmp;
    last_type++;
    usrp++;

    usrtab[usrp].symp = symp;
    usrtab[usrp].flg = 0;
    usrtab[usrp].extc = 0;
    usrtab[usrp].ext = NULL;
    RA = symp;
    RB = cscope;
    cscope = last_type;
    fprintf(dec, "typedef struct {\n");
    addh(typtab, n, last_type);
    symf |= sf_INSIDE_TYPE;
}

void sym_leave_type()
{
    int i;
    if (usrtab[usrp].symp == symp) {
	msgrep(mWRN, "Empty type block; discarded");
	usrp--;
	return;
    }
    usrtab[usrp].agr = 0;
    for (i = usrtab[usrp].symp + 1; i <= symp; i++) {
	if (symtab[i].ttp.cl <= cARRAY)
	    usrtab[usrp].agr += sym_sizes(symtab[i].ttp.tp);
    }
    for (i = 0; i < usrtab[usrp].extc; i++)
	fprintf(dec, "_fT_%s _dummy_%d;\n",
		hash_by_val(typtab, usrtab[usrp].ext[i]), i);


    cscope = RB;
    symf &= ~sf_INSIDE_TYPE;
    fprintf(dec, "} _fT_%s;\n", hash_by_val(typtab, last_type));
    dec = lfh;
    if (usrtab[usrp].flg & (tp_WINIT | tp_WKILL))
	*spref = 0;
}


void symtab_dec(long p)
{
    int i, d;
    if (symf & sf_NODEC)
	return;
    switch (symtab[p].ttp.cl) {
    case cSCALAR:{
	    print_cmpl_type(dec, symtab[p].ttp.tp);
	    if (symtab[p].ttp.tp >= tMAX)
		fprintf(dec, "*");
	    fprintf(dec, " _fs_%s", symtab[p].name);
	    print_init_val(dec, p);
	    break;
	}
    case cARRAY:{
	    d = FDIMS(p);
	    if (FF(p) & af_DYNAMIC) {
		fprintf(dec, "struct FC_arr _fa_%s;\n", symtab[p].name);
		fprintf(out, "FC_arr_c(&_fa_%s,%d,sizeof(", symtab[p].name,
			FTP(p));
		print_cmpl_type(out, FTP(p));
		fprintf(out, "),%d", d);
		for (i = 0; i < d; i++)
		    fprintf(out, ",%s", DNTH(p, i));
		fprintf(out, ");\n");
	    } else {
		print_cmpl_type(dec, FTP(p));
		if (FTP(p) >= tMAX)
		    fprintf(dec, "*");
		fprintf(dec, " _fa_%s", symtab[p].name);
		for (i = 0; i < d; i++)
		    fprintf(dec, "[%ld]", SNTH(p, i));
		fprintf(dec, ";\n");
	    }
	    break;
	}
    }
}




long symtab_add_sym(char *name, int tp, int cl, int scope)
{
    long i;

    if (symf & sf_NO_AUTO_DEC) {
	if (symf & sf_COMPLX_NAME)
	    msgrep(mERR, "no such a member - <%s> in type %s", name,
		   hash_by_val(typtab, tp));
	else
	    msgrep(mERR, "can't auto-declare this name - %s\n", name);
	return -1;
    }
    if ((i = symtab_look(name, scope)) > 0 && !(symf & sf_ALLOW_REDEC))
	msgrep(mERR, "name redeclaration - `%s'", name);
    symp++;
    symtab[symp].flg = 0;
    symtab[symp].name = lcase(strdup(name));
    symtab[symp].ttp.cl = cl;
    symtab[symp].ttp.tp = tp;
    if (!_pub)
	symtab[symp].flg |= nf_PROTECTED;
    // printf("Add_sym: ? TR symf=%d\n",symf & (sf_INSIDE_TYPE | sf_BEGINCD));
    if ((symf & sf_INSIDE_TYPE) && (symf & sf_BEGINCD))
	symtab[symp].scope = -1;
    else
	symtab[symp].scope = scope;
    symtab[symp].adata = (void *) 0;
    if (symf & sf_INSIDE_LET)
	return symp;
    symtab_dec(symp);

    // use constructor for user types
    if (tp >= tMAX && (usrtab[tp - tMAX].flg & tp_WINIT)
	&& !(symf & sf_INSIDE_TYPE))
	use_constructor(name, tp, 0);

    return symp;
}

long symtab_add_sym_full(char *name, int tp, int cl, int scope, void *d)
{
    if (symf & sf_NO_AUTO_DEC) {
	if (symf & sf_COMPLX_NAME)
	    msgrep(mERR, "no such a member - <%s> in type %s", name,
		   hash_by_val(typtab, tp));
	else
	    msgrep(mERR, "can't auto-declare this name - %s\n", name);
	return -1;
    }
    if (symtab_look(name, scope) && !(symf & sf_ALLOW_REDEC))
	msgrep(mERR, "name redeclaration - `%s'", name);
    symp++;
    symtab[symp].flg = 0;
    symtab[symp].name = lcase(strdup(name));
    symtab[symp].ttp.cl = cl;
    symtab[symp].ttp.tp = tp;
    if (!_pub)
	symtab[symp].flg |= nf_PROTECTED;
    if ((symf & sf_INSIDE_TYPE) && (symf & sf_BEGINCD))
	symtab[symp].scope = -1;
    else
	symtab[symp].scope = scope;
    symtab[symp].adata = d;
    symtab_dec(symp);
    if (tp >= tMAX && (usrtab[tp - tMAX].flg & tp_WINIT)
	&& !(symf & sf_INSIDE_TYPE))
	use_constructor(name, tp, 0);
    return symp;
}

void symtab_set_sym_full(long s, char *name, etp t, int scope, void *d)
{
    if (symtab[s].name)
	free(symtab[s].name);
    symtab[s].flg = 0;
    symtab[s].name = lcase(strdup(name));
    symtab[s].ttp = t;
    symtab[s].scope = scope;
    symtab[s].adata = d;
}


void symtab_add_sym_arr(char *name, int tp, int cl, int scope, int dims,
			void *adata)
{
    if (symtab_look(name, scope) && !(symf & sf_ALLOW_REDEC))
	msgrep(mERR, "Name redeclaration - `%s'", name);
    symp++;
    symtab[symp].flg = 0;
    symtab[symp].name = lcase(strdup(name));
    symtab[symp].ttp.cl = cl;
    symtab[symp].ttp.tp =
	alloc_arr_type(tp, dims,
		       (((arr_d *) adata)->flg & af_DYNAMIC) ? 1 : 0, 0);
    if ((symf & sf_INSIDE_TYPE) && (symf & sf_BEGINCD))
	symtab[symp].scope = -1;
    else
	symtab[symp].scope = scope;
    symtab[symp].adata = adata;
    symtab_dec(symp);
    if (tp >= tMAX && (usrtab[tp - tMAX].flg & tp_WINIT)) {
	if ((((arr_d *) (symtab[symp].adata))->flg) & af_DYNAMIC)
	    fprintf(out, "FC_arr_init2(&_fa_%s,_%s_ff_%ld_init);", name,
		    hash_by_val(typtab, tp), look_sub("init", tp, 0));
	else
	    fprintf(out,
		    "FC_arr_init1((void*)_fa_%s,_%s_ff_%ld_init,%ld,sizeof(_fT_%s));",
		    name, hash_by_val(typtab, tp), look_sub("init", tp, 0),
		    arr_static_size(symp), hash_by_val(typtab, tp));
    }
}



ir_etp symtab_get_ir(char *name, int scope, int cc)
{
    int i;
    ir_etp t;
    for (i = 0; i <= symp; i++)
	if (!strcasecmp(symtab[i].name, name)) {
	    t.t = symtab[i].ttp;
	    t.adata = symtab[i].adata;
	    return t;
	}
    symtab_add_sym(name, tINT, cc, scope);
    t.t = symtab[symp].ttp;
    t.adata = symtab[symp].adata;
    return t;
}


void symtab_add_sym_chk(char *name, int tp, int scope)
{
    etp t;
    arr_d *d;
    char *nm;
    int tl = 0, i = 0, is_c = 1, k = 0;
    t = popbt(stk0);
    if (t.tp == cSCALAR) {
	symtab_add_sym(name, tp, cSCALAR, scope);
	return;
    }

    tl += strlen(name);
    pushb(stk1, name, t);

    do {

	if (t.cl == cSCALAR)
	    is_c = 0;
	else {
	    if (t.cl != cCONST)
		msgrep(mERR,
		       "const/scalar class expected for subscript of %s",
		       name);}

	nm = popb(stk0, &t);
	tl += strlen(nm);
	i++;
	pushb(stk1, nm, t);
    } while (t.tp != cARRAY);

    name = popb(stk1, &t);
    dims = i;
    d = (arr_d *) malloc(sizeof(arr_d));
/* if (is_c) printf("constant\n");
      else printf("non-constant\n"); 
*/
    d->flg = 0;
    if (is_c)
	d->d.dim_v = (long *) malloc((i + 1) * sizeof(long));
    else {
	d->d.dim_e = (char **) malloc((i + 1) * sizeof(char *));
	d->flg |= af_DYNAMIC;
    }

    do {
	nm = popb(stk1, &t);
	if (is_c)
	    d->d.dim_v[k] = atoi(nm);
	else
	    d->d.dim_e[k] = strdup(nm);
	k++;
	i--;
    } while (i);
    symtab_add_sym_arr(name, tp, cARRAY, scope, dims, d);
}

etp symtab_add_sym_chk_ret(char *name, int tp, int scope)
{
    symtab_add_sym_chk(name, tp, scope);
    sprintf(stk0->cbuf, "_fs_%s", symtab[symp].name);
    stk0->cbuf += strlen(stk0->cbuf) + 1;
    return symtab[symp].ttp;
}


etp sym_get_sym_arr()
{
    etp t;
    int tl = 0, i = 0, i2 = 1, j, flg;
    long l;
    char *nm, *mbf, *nm2;

    do {
	nm = popb(stk0, &t);
	tl += strlen(nm);
	i++;
	pushb(stk1, nm, t);
    } while ((t.tp != cARRAY) || (t.cl != cSPEC));
    nm = popb(stk1, &t);

    if (_is_built_in(nm))
	return cmpl_built_in(i - 1);

    if (!(j = symtab_look(nm, 0)))
	if (!(j = symtab_look(nm, cscope))) {
	    msgrep(mHLT, "%s:%s/%d - declared symbol expected!\n", nm,
		   __FILE__, __LINE__);
	    exit(0);
	}
    j--;

    if (symtab[j].ttp.cl == cFUN) {
	nm = strdup(nm);
	if (_mc) {
	    nm2 = popb(stk0, &t);
	    pushb(stk1, nm2, t);
	}
	t = fune_call(nm, i - 1, j);
	free(nm);
	addhr(ctx, "", ctx_PART_NAME);
	return t;
    }

    if (symf & sf_EMPTY_ELIST)
	msgrep(mERR, "%s/%d - explicit range expected for %s", __FILE__,
	       __LINE__, nm);
    symf &= ~sf_LEFT_SIDENM;
    mbf =
	(char *) malloc((tl + i * 2 + 108 + strlen(spref)) * sizeof(char));
    flg = ((arr_d *) (symtab[j].adata))->flg;

    if (symtab[j].scope != 1 && symtab[j].scope != -1
	&& !(symf & sf_COMPLX_NAME)) {
	sprintf(mbf, "%s", spref);
    } else
	*mbf = 0;

    nm2 = lcase(strdup(nm));
    if (flg & af_DYNAMIC) {
	if (!req_adr)
	    sprintf(mbf, "%s(*(", mbf);
	else
	    sprintf(mbf, "%s((", mbf);
	mbf = sprint_cmpl_type(mbf, FTP(j));
	sprintf(mbf, "%s*)((char*)_fa_%s.data", mbf, nm2);
    } else
	sprintf(mbf, "%s_fa_%s", mbf, nm2);
    addhr(ctx, nm, ctx_PART_NAME);
    nm = popb(stk1, &t);
    i -= 1;
    if (DIMS(symtab[j].ttp.tp) != i) {
	if (flg & af_DYNAMIC)
	    msgrep(mWRN,
		   "possibly wrong number of subscripts for dynamic array %s (%d used, %d expected)",
		   symtab[j].name, i, DIMS(symtab[j].ttp.tp));
	else
	    msgrep(mERR, "for array %s got %d subscripts, expected %d",
		   symtab[j].name, i, DIMS(symtab[j].ttp.tp));
    }
    if (!(flg & af_DYNAMIC) && t.cl == cCONST) {
	l = atol(nm);
	if (l < 0 || (l > ((arr_d *) (symtab[j].adata))->d.dim_v[i2]))
	    msgrep(mWRN, "%d range subscript out of limit for %s", i2 + 1,
		   nm2);
    };
    if (flg & af_DYNAMIC)
	sprintf(mbf, "%s+(%s", mbf, nm);
    else
	sprintf(mbf, "%s[%s]", mbf, nm);
    nm = popb(stk1, &t);
    i--;

    while (i) {
	if (!(flg & af_DYNAMIC) && t.cl == cCONST) {
	    l = atol(nm);
	    if (l < 0 || (l > ((arr_d *) (symtab[j].adata))->d.dim_v[i2]))
		msgrep(mWRN, "%d range subscript out of limit for %s",
		       i2 + 1, nm2);
	};
	if (flg & af_DYNAMIC)
	    sprintf(mbf, "%s+%s*_fa_%s.dm[%d]", mbf, nm, nm2, i2 - 1);
	else
	    sprintf(mbf, "%s[%s]", mbf, nm);
	nm = popb(stk1, &t);
	i2++;
	i--;
    };
    if (flg & af_DYNAMIC) {
	sprintf(mbf, "%s)*sizeof(", mbf);
	mbf = sprint_cmpl_type(mbf, FTP(j));
	sprintf(mbf, "%s)))", mbf);
    }
    t.tp = arp[symtab[j].ttp.tp].tp;
    t.cl = cSCALAR;
    pushb(stk0, mbf, t);
    free(mbf);
    return t;
}

etp symtab_get_sym(char *name)
{
    int i;
    char *my_name, *nm;
    etp t;
    addhr(ctx, name, ctx_PART_NAME);
    my_name = strdup(popbe(stk0));
    name = my_name;
    lcase(name);
    sprintf(stk0->cbuf, "");

    for (i = 0; i <= symp; i++) {
	if (symtab_gate(name, cscope, i)) {

	    if (symtab[i].ttp.cl == cFUN) {
		if (_mc) {
		    nm = popb(stk0, &t);
		    pushb(stk1, nm, t);
		} else
		    break;
		if (!((symf & sf_INSIDE_TYPE)
		      && symtab[i].scope == last_type)
		    && (symtab[i].flg & nf_PROTECTED))
		    msgrep(mERR,
			   "Can't access protected class member - %s",
			   name);

		t = fune_call(name, 0, i);
		t = popbt(stk0);
		free(name);
		addhr(ctx, "", ctx_PART_NAME);
		return t;
	    }

	    if (symtab[i].scope != 1 && symtab[i].scope != -1
		&& !(symf & sf_COMPLX_NAME)) sprintf(stk0->cbuf, "%s",
						     spref);
	    if (symtab[i].flg & nf_BYREF)
		sprintf(stk0->cbuf, "%s*", stk0->cbuf);
	    if (req_adr)
		sprintf(stk0->cbuf, "%s&_f%c_%s", stk0->cbuf,
			cmpl_pref(symtab[i].ttp.cl), name);
	    else
		sprintf(stk0->cbuf, "%s_f%c_%s", stk0->cbuf,
			cmpl_pref(symtab[i].ttp.cl), name);
	    stk0->cbuf += strlen(stk0->cbuf) + 1;
	    free(name);
	    symf &= ~sf_INSIDE_LET;
	    return symtab[i].ttp;
	}
    }
    msgrep(mNOT, "symbol %s not declared; forcing declaration", name);
    if ((symf & sf_INSIDE_TYPE) && (symf & sf_BEGINCD))
	symtab_add_sym(name, c_tp, c_cl, -1);
    else
	symtab_add_sym(name, c_tp, c_cl, cscope);

    if (req_adr)
	sprintf(stk0->cbuf, "&_fs_%s", name);
    else
	sprintf(stk0->cbuf, "_fs_%s", name);
    stk0->cbuf += strlen(stk0->cbuf) + 1;
    free(name);
    return SC_INT;
}


char *lcase(char *s)
{
    register char *t = s;
    while (*s) {
	*s = tolower(*s);
	s++;
    }
    return t;
}

void symtab_show(char *ff)
{
    int i, t;
    FILE *f;
    f = fopen(ff, "w+");
    fprintf(f, "Sym Tab: total %ld entries:\n\n", symp + 1);
    for (i = 0; i <= symp; i++) {
	if (symtab[i].scope >= tMAX)
	    fprintf(f, "\t%20s(%4d):%s - ",
		    hash_by_val(typtab, symtab[i].scope), symtab[i].scope,
		    symtab[i].name);
	else
	    fprintf(f, "\t%26d:%s - ", symtab[i].scope, symtab[i].name);
	if (symtab[i].flg & nf_BYREF)
	    fprintf(f, "pointer to ");
	switch (symtab[i].ttp.cl) {
	case cCONST:{
		fprintf(f, "const");
		t = symtab[i].ttp.tp;
		break;
	    }
	case cSCALAR:{
		fprintf(f, "scalar");
		t = symtab[i].ttp.tp;
		break;
	    }
	case cARRAY:{
		fprintf(f, "array");
		t = arp[symtab[i].ttp.tp].tp;
		fprintf(f, " of %d dimensions",
			arp[symtab[i].ttp.tp].dims);
		break;
	    }
	case cFUN:{
		fprintf(f, "function");
		t = symtab[i].ttp.tp;
		break;
	    }
	default:{
		fprintf(f, "*?class?*");
		break;
		t = stANY;
	    }
	}
	switch (t) {
	case stANY:{
		fprintf(f, " any");
		break;
	    }
	case tINT:{
		fprintf(f, " int");
		break;
	    }
	case tDBL:{
		fprintf(f, " double");
		break;
	    }
	default:{
		fprintf(f, " %s", hash_by_val(typtab, t));
		break;
	    }
	}
	if (symtab[i].flg & nf_PROTECTED)
	    fprintf(f, " (protected)");
	fprintf(f, "\n");
    }
    if (last_type >= tMAX) {
	fprintf(f, "\t\nClasses:");
	while (last_type >= tMAX) {
	    fprintf(f, "\n\t\t%d - %s ", last_type,
		    hash_by_val(typtab, last_type));
	    if (usrtab[last_type - tMAX].extc) {
		fprintf(f, " extends ");
		while (usrtab[last_type - tMAX].extc) {
		    usrtab[last_type - tMAX].extc--;
		    fprintf(f, "%s;",
			    hash_by_val(typtab,
					usrtab[last_type -
					       tMAX].ext[usrtab[last_type -
								tMAX].
							 extc]));
		}
	    }
	    last_type--;
	}
    }
    fprintf(f, "\n\nCS:%d\n", cscope);
    fclose(f);
}

etp metp(int t, int c)
{
    etp n;
    n.tp = t;
    n.cl = c;
    return n;
}

char cmpl_pref(int cl)
{
    switch (cl) {
    case cCONST:
	return 'c';
    case cSCALAR:
	return 's';
    case cARRAY:
	return 'a';
    case cFUN:
	return 'f';
    default:
	return '_';
    }
}

void print_cmpl_type(FILE * wt, int t)
{
    switch (t) {
    case tINT:
	fprintf(wt, "int");
	break;
    case tLONG:
	fprintf(wt, "long int");
	break;
    case tWORD:
	fprintf(wt, "unsigned int");
	break;
    case tDWORD:
	fprintf(wt, "unsigned long int");
	break;
    case tREAL:
	fprintf(wt, "float");
	break;
    case tDBL:
	fprintf(wt, "double");
	break;
    case tSTR:
	fprintf(wt, "char*");
	break;
    default:
	fprintf(wt, "_fT_%s", hash_by_val(typtab, t));
    }

}

char *sprint_cmpl_type(char *wt, int t)
{
    switch (t) {
    case tINT:
	sprintf(wt, "%sshort int", wt);
	break;
    case tLONG:
	sprintf(wt, "%slong int", wt);
	break;
    case tWORD:
	sprintf(wt, "%sunsigned short int", wt);
	break;
    case tDWORD:
	sprintf(wt, "%sunsigned long int", wt);
	break;
    case tREAL:
	sprintf(wt, "%sfloat ", wt);
	break;
    case tDBL:
	sprintf(wt, "%sdouble", wt);
	break;
    case tSTR:
	sprintf(wt, "%schar*", wt);
	break;
    default:
	sprintf(wt, "%s_fT_%s", wt, hash_by_val(typtab, t));
    }
    return wt;
}

void print_init_val(FILE * wt, long p)
{
    int c;
    if (cscope > 0) {
	fprintf(wt, ";\n");
	fprintf(out, "_fs_%s", symtab[p].name);
	c = cscope;
	cscope = 0;
	print_init_val(out, p);
	cscope = c;
	return;
    }

    switch (symtab[p].ttp.tp) {
    case tINT:
    case tLONG:
    case tWORD:
    case tDWORD:
	fprintf(wt, "=0;\n");
	break;
    case tDBL:
    case tREAL:
	fprintf(wt, "=0.0;\n");
	break;
    case tSTR:
	fprintf(wt, "=FCes;\n");
	break;
    default:
	if (symtab[p].ttp.tp >= tMAX) {
	    fprintf(wt, ";\n");
	    fprintf(out, "_fs_%s=(_fT_%s*) ", symtab[p].name,
		    hash_by_val(typtab, symtab[p].ttp.tp));
	    if (cscope != 0)
		fprintf(out, "mmalloc");
	    else
		fprintf(out, "m2malloc");
	    fprintf(out, "(sizeof(_fT_%s));\n",
		    hash_by_val(typtab, symtab[p].ttp.tp));
	} else
	    fprintf(wt, ";\n");
    }
}

void print_ini_val(FILE * wt, int p)
{
    switch (p) {
    case tINT:
    case tLONG:
    case tWORD:
    case tDWORD:
	fprintf(wt, "=0;\n");
	break;
    case tDBL:
    case tREAL:
	fprintf(wt, "=0.0;\n");
	break;
    case tSTR:
	fprintf(wt, "=FCes;\n");
	break;
    default:
	if (p >= tMAX)
	    fprintf(wt, "=(_fT_%s*) mmalloc(sizeof(_fT_%s));\n",
		    hash_by_val(typtab, p), hash_by_val(typtab, p));
	else
	    fprintf(wt, ";\n");
    }

}

long arr_static_size(long j)
{
    int i, td, ts = 1;
    td = DIMS(symtab[j].ttp.tp);
    for (i = 0; i < td; i++)
	ts *= ((arr_d *) (symtab[j].adata))->d.dim_v[i];
    return ts;
}

void use_constr()
{
    use_constructor(symtab[symp].name, symtab[symp].ttp.tp, _nc);
}

void use_constructor(char *name, int tp, int pc)
{
    char *t;
    int os = cscope, jj;
    jj = symtab_look("init", tp) - 1;
    if (jj == -1)
	return;
    cscope = tp;
    _mc = 1;
    t = (char *) malloc(strlen(name) + 5);
    sprintf(t, "_fs_%s", name);
    pushb(stk1, t, metp(tp, cSCALAR));
    fune_call("init", pc, jj);
    call_write();
    free(t);
    cscope = os;
    _mc = 0;
}

void analyze_me()
{
    if (symf & sf_INSIDE_TYPE) {
	sprintf(stk0->cbuf, "_Xfs_x");
	stk0->cbuf += strlen(stk0->cbuf) + 1;
	pushbt(stk0, metp(last_type, cSCALAR));
    } else
	msgrep(mERR, "Value of ME is undefined in this context");
}

long sym_sizes(int t)
{
    switch (t) {
    case tINT:
	return sizeof(short int);
    case tLONG:
	return sizeof(long int);
    case tWORD:
	return sizeof(unsigned short int);
    case tDWORD:
	return sizeof(unsigned long int);
    case tREAL:
	return sizeof(float);
    case tDBL:
	return sizeof(double);
    case tSTR:
	return sizeof(char *);
    }
    if (t >= tMAX && t <= last_type) {
	return usrtab[t - tMAX].agr;
    }
    msgrep(mHLT, "Internal error: invalid type request - %d", t);
}
