/*
This file is a part of Lemick Basic (C) Copyright Alex Iliasov 2001.
function declaration and call
also function overloading implemented here
*/

#include <stdlib.h>
#include <string.h>
#include "glb.h"
#include "sym.h"
#include "cbufs.h"
#include "igo.h"
FILE *old_out, *old_dec;

extern usert *usrtab;
extern hash *typtab;
extern hash *fileh, *sfileh, *tfileh;
extern symt *symtab;
extern long usrp, symp;
FILE *reset(char *n, FILE * f);


struct _fun_t {
    char flg;			//function flags
    char *tflg;			//flags for each argument
    etp *typ;			//types and classes of arguments
    char **nom;			//names of arguments
    char **opt;			//list of OPTIONAL initialazers
    int pc, pceo;		//number of args. (incl./excl. optional)
};
typedef struct _fun_t fun_t;

struct ir_etp_ {
    etp t;
    void *adata;
};

typedef struct ir_etp_ ir_etp;

union dims_t {
    char **dim_e;		// range subscripts as compiled to C expressions
    long *dim_v;		// or numbers, if array is static 
};

struct arr_d_ {
    int flg;			// array-specific flags 
    union dims_t d;
};

typedef struct arr_d_ arr_d;


enum func_flags {
    ff_FUNCTION = 1 << 0,	//returns value
    ff_EXTERN = 1 << 1,		//body of function is somewhere in library
    //thus don't change name
    ff_THREAD = 1 << 2,		//compile as thread
    ff_DYNAMIC = 1 << 3,	//function body is in DSO
    ff_PROTO = 1 << 4,		//function body goes later (in this file)
    ff_PUBLIC = 1 << 5,		//method is visible outside it's class
};

/*    //these reside in sym.h
enum name_flags {
     nf_BYVAL    = 1<<0,
     nf_BYREF    = 1<<1,
     nf_OPTIONAL = 1<<3,
     nf_EXTYPE   = 1<<4,
     nf_ARTYPE   = 1<<5
 };
*/

enum fundec_flags {
    fdf_WRITE = 1 << 0,
    fdf_FUNCTION = 1 << 1,
    fdf_DECLARE = 1 << 2,
    fdf_CD = 1 << 3,		//is a method (inside TYPE or CLASS)   
    fdf_DYNAMIC = 1 << 4,	//function body is in DSO
    fdf_EXTERN = 1 << 5,	//body of function is somewhere in library
    fdf_THREAD = 1 << 6		//compile as thread
};


fun_t *funp;
long func = -1;
long funmax = 100;
char *cfn = NULL;
int _oac, lu;


void write_fun_dec(long sp, char *nm, int i, int n);
long look_overloads(long j, long p);
int penalty(etp a, etp b);
etp fune_call_nov(const char *nm, int k, int j);
long look_strict(char *nm, int scope, fun_t * fj);


void init_fn()
{
    funp = (fun_t *) malloc(funmax * sizeof(fun_t));
}

void free_fn()
{
    free(funp);
}


void enter_fun_dec(int nc, int flg)
{
    int i, oi, _symf, j;
    long sp;
    char *nm;
    etp t, rt;
    arr_d *d;
    lu = 0;
    func++;
    if (nc) {
	funp[func].tflg = (char *) malloc(sizeof(char) * nc);
	funp[func].typ = (etp *) malloc(sizeof(etp) * nc);
	funp[func].nom = (char **) malloc(sizeof(char *) * nc);
    } else {
	funp[func].tflg = NULL;
	funp[func].typ = NULL;
	funp[func].nom = NULL;
    }
    funp[func].pc = nc;
    funp[func].pceo = nc;
    funp[func].flg |= flg;
    symtab_add_sym("", tINT, cCONST, -1);
    sp = symp;
    if (flg & fdf_FUNCTION) {
	popb(stk0, &rt);
    }
    if (_oac)
	funp[func].opt = (char **) malloc(sizeof(char *) * _oac);
    i = nc;
    oi = nc;
    _symf = symf;
    symf |= sf_NODEC;
    if (i > 0)
	do {
	    funp[func].tflg[i - 1] = pop(far);
	    funp[func].nom[i - 1] =
		strdup(popb(stk1, &funp[func].typ[i - 1]));
	    if (funp[func].tflg[i - 1] & nf_OPTIONAL) {
		if (funp[func].tflg[i - 1] & nf_BYREF)
		    msgrep(mERR,
			   "OPTIONAL argument can't be passed BYREF");
		funp[func].pceo = nc - i;
		funp[func].opt[oi - 1] = strdup(popb(stk0, &t));
		if (t.cl != cCONST)
		    msgrep(mERR,
			   "constant expression required for OPTIONAL argument's default value");
		oi--;
	    }

	    if (funp[func].typ[i - 1].cl == cARRAY) {
		d = (arr_d *) calloc(sizeof(arr_d), 1);
		d->d.dim_v =
		    (long *) calloc(DIMS(funp[func].typ[i - 1].tp),
				    sizeof(long));
		if (SF(funp[func].typ[i - 1].tp)) {
		    d->flg |= af_DYNAMIC;
		}
		for (j = 1; j < DIMS(funp[func].typ[i - 1].tp); j++) {
		    d->d.dim_v[j] = atol(popb(stk0, &t));
		}
		symtab_add_sym_arr(funp[func].nom[i - 1],
				   TP(funp[func].typ[i - 1].tp), cARRAY,
				   -1, DIMS(funp[func].typ[i - 1].tp), d);
	    } else {
		symtab_add_sym(funp[func].nom[i - 1],
			       funp[func].typ[i - 1].tp,
			       funp[func].typ[i - 1].cl, -1);
	    }

	    if (funp[func].tflg[i - 1] & nf_BYREF) {
		symtab_set_flg(symp, nf_BYREF);
		printf("set BYREF to %s?%s (%d)\n", funp[func].nom[i - 1],
		       symtab_name(symp), symp);
	    }
	    i--;
	} while (i > 0);

    symf = _symf;
    nm = popb(stk1, &t);
    cfn = strdup(nm);
    if (flg & fdf_FUNCTION)
	i = rt.tp;
    else
	i = tVOID;

    symf |= sf_ALLOW_REDEC;
    j = look_strict(nm, cscope, &funp[func]);
    if (j) {
	j--;
	if (!(((fun_t *) symtab_adata(j))->flg & fdf_DECLARE))
	    msgrep(mERR, "Function redeclaration - %s", nm);
	((fun_t *) symtab_adata(j))->flg &= ~fdf_DECLARE;
    } else {
	symtab_set_sym_full(sp, nm, metp(i, cFUN), cscope, &funp[func]);
	j = sp;
    }

    if (flg & fdf_EXTERN)
	symtab_set_flg(sp, nf_EXTERN);

    symf &= ~sf_ALLOW_REDEC;

    if (flg & fdf_CD) {
	usrtab_sf(usrp, tp_WINIT);
	if (!strcasecmp(cfn, "init"))
	    usrtab_sf(usrp, tp_WINIT);
	else {
	    if (!strcasecmp(cfn, "kill"))
		usrtab_sf(usrp, tp_WKILL);
	}
	*spref = 0;
	sprintf(spref, "_Xfs_x->");
	symf |= sf_BEGINCD;
    };

    if (!(flg & fdf_DECLARE)) {
	push(ges, cscope);
	cscope = -1;
    }

    if (flg & fdf_WRITE)
	write_fun_dec(j, nm, i, func);
    if (flg & fdf_DECLARE) {
	delete_scope(-1);
	return;
    }
    tfileh = fileh;
    fileh = sfileh;
    freeh(fileh);
    old_out = out;
    old_dec = dec;
    out = tmp1;
    dec = tmp2;
}



void leave_fun_dec()
{
    cscope = pop(ges);
    if (lu)
	fprintf(out, "_llf_%d:\n", func);
    fileh = tfileh;
    if (funp[func].flg & fdf_CD) {
	if (funp[func].flg & fdf_FUNCTION)
	    fprintf(out, "return _fs_%s;\n", cfn);
	fprintf(out, "}\n");
	out = old_out;
	dec = old_dec;
	f2f(sub, tmp2);
	reset(ntmp2, tmp2);
	f2f(sub, tmp1);
	reset(ntmp1, tmp1);
	symf &= ~sf_BEGINCD;
	delete_scope(-1);
	return;
    }

    delete_scope(-1);
    out = old_out;
    dec = old_dec;
    cscope = 0;

    f2f(sub, tmp2);
    reset(ntmp2, tmp2);


    f2f(sub, tmp1);
    reset(ntmp1, tmp1);

    if (funp[func].flg & fdf_FUNCTION)
	fprintf(sub, "return _fs_%s;\n", cfn);
    fprintf(sub, "}\n");

}

void write_fun_dec(long sp, char *nm, int i, int n)
{
    int k = 0, j, s;
    if (funp[n].flg & fdf_FUNCTION)
	print_cmpl_type(sub, i);
    else
	fprintf(sub, "void");
    if (i >= tMAX)
	fprintf(sub, "*");
    fprintf(sub, " ");
    if (!(funp[n].flg & fdf_EXTERN)) {
	if ((funp[n].flg & fdf_CD))
	    fprintf(sub, "_%s_ff_%ld_", hash_by_val(typtab, last_type),
		    sp);
	else
	    fprintf(sub, "_ff_%ld_", sp);
    }
    fprintf(sub, "%s(", lcase(nm));

    // write header for constr/destr
    if (funp[n].flg & fdf_CD)
	fprintf(sub, "_fT_%s* _Xfs_x", hash_by_val(typtab, last_type));

    if (funp[n].pc > 0) {
	if (funp[n].flg & fdf_CD)
	    fprintf(sub, ",");
	if (funp[n].typ[0].cl == cSCALAR) {
	    print_cmpl_type(sub, funp[n].typ[0].tp);
	    if (funp[n].tflg[0] & nf_BYREF || (funp[n].typ[0].tp >= tMAX))
		fprintf(sub, "*");
	    fprintf(sub, " _fs_%s", funp[n].nom[0]);
	} else {
	    s = symtab_look(funp[n].nom[0], cscope) - 1;
	    if (FF(s) & af_DYNAMIC) {
		fprintf(sub, "struct FC_arr _fa_%s", funp[n].nom[0]);
	    } else {
		print_cmpl_type(sub, TP(funp[n].typ[0].tp));
		fprintf(sub, " _fa_%s[]", funp[n].nom[0]);
		for (j = 1; j < DIMS(funp[n].typ[0].tp); j++)
		    fprintf(sub, "[%ld]", SNTH(s, j));
	    }
	}
	k++;
    }

    while (k < funp[n].pc) {
	fprintf(sub, ",");
	if (funp[n].typ[k].cl == cSCALAR) {
	    print_cmpl_type(sub, funp[n].typ[k].tp);
	    if (funp[n].tflg[k] & nf_BYREF || (funp[n].typ[k].tp >= tMAX))
		fprintf(sub, "*");
	    fprintf(sub, " _fs_%s", funp[n].nom[k]);
	} else {
	    s = symtab_look(funp[n].nom[k], cscope) - 1;
	    if (FF(s) & af_DYNAMIC) {
		fprintf(sub, "struct FC_arr _fa_%s", funp[n].nom[k]);
	    } else {
		print_cmpl_type(sub, TP(funp[n].typ[k].tp));
		fprintf(sub, " _fa_%s[]", funp[n].nom[k]);
		for (j = 1; j < DIMS(funp[n].typ[k].tp); j++)
		    fprintf(sub, "[%ld]", SNTH(s, j));
	    }
	}
	k++;
    }


    if (funp[n].flg & fdf_DECLARE) {
	fprintf(sub, ");\n");
	return;
    }
    fprintf(sub, ") {");

    if (funp[func].flg & fdf_FUNCTION) {
	symf |= sf_NODEC;
	symtab_add_sym(nm, i, cSCALAR, -1);
	symf &= ~sf_NODEC;
	fprintf(sub, "\n");
	print_cmpl_type(sub, i);
	if (i >= tMAX)
	    fprintf(sub, "*");
	fprintf(sub, " _fs_%s", nm);
	print_ini_val(sub, i);
    }

    if ((funp[n].flg & fdf_CD) && (!strcasecmp(nm, "init")))
	write_subtypes_init(sub, last_type);

    symf &= ~sf_NODEC;
}



void call_fun()
{
    int i, ic = 0, sc;
    etp t;
    char *nm;
    fun_t *ft;
    fileh = tfileh;

    nm = popb(stk0, &t);


    while (t.cl != cSPEC) {
	pushb(stk1, nm, t);
	nm = popb(stk0, &t);
	ic++;
    }
    i = symtab_look(nm, cscope);
    if (!i) {
	printf("sub-routine <%s> not found\n", nm);
	exit(0);
    }
    i--;
    ft = (fun_t *) symtab_adata(i);
    t.cl = symtab_class(i);


    if (t.cl != cFUN) {
	printf("<%s> is not a function\n", nm);
	exit(0);
    }

    fprintf(out, "_ff_%d_%s(", i, nm);
    sc = ic;
    if (ft->pceo > sc)
	msgrep(mERR, "Too few arguments for %s subroutine", nm);
    if (ft->pc < sc)
	msgrep(mERR, "Too many arguments for %s subroutine", nm);

    if (ic > 0) {
	nm = popb(stk1, &t);
	fprintf(out, "%s", nm);
	ic--;
    }

    while (ic > 0) {
	nm = popb(stk1, &t);
	fprintf(out, ",%s", nm);
	ic--;
    }
    if (ft->pc > sc) {
	if (!sc) {
	    fprintf(out, "%s", ft->opt[sc - ft->pceo]);
	    sc++;
	}
	while (ft->pc > sc) {
	    fprintf(out, ",%s", ft->opt[sc - ft->pc]);
	    sc++;
	}
    }

    fprintf(out, ");");
}


etp fune_call_plain(const char *nm, int k, int scope)
{
    return fune_call_nov(nm, k, symtab_look(nm, scope) - 1);
}

etp fune_call_nov(const char *nm, int k, int j)
{
    int i, jj;
    etp t, tt;
    fun_t *ft;
    char *a;

    t = metp(symtab_type(j), cSCALAR);
    ft = (fun_t *) symtab_adata(j);


    if (!symtab_test_flg(j, nf_EXTERN)) {
	if (_mc)
	    sprintf(stk0->cbuf, "_%s_ff_%d_",
		    hash_by_val(typtab, symtab_scope(j)), j);
	else
	    sprintf(stk0->cbuf, "_ff_%d_", j);
	sprintf(stk0->cbuf, "%s%s(", stk0->cbuf, nm);
    } else
	sprintf(stk0->cbuf, "%s(", nm);

    if (_mc) {
	a = popb(stk1, &tt);
	i = symtab_scope(j);
	if (tt.tp != i) {
	    sprintf(stk0->cbuf, "%s(_fT_%s*)(%s", stk0->cbuf,
		    hash_by_val(typtab, i), a);
	    sprintf(stk0->cbuf, "%s+%ld", stk0->cbuf,
		    usrtab[tt.tp - tMAX].agr);
	    jj = usrtab[tt.tp - tMAX].extc;
	    i = 0;
	    while (i < jj) {
		if (usrtab[tt.tp - tMAX].ext[i] == symtab_scope(j))
		    break;
		sprintf(stk0->cbuf, "%s+%ld", stk0->cbuf,
			usrtab[usrtab[tt.tp - tMAX].ext[i] - tMAX].agr);
		i++;
	    }
	    sprintf(stk0->cbuf, "%s)", stk0->cbuf);
	} else
	    sprintf(stk0->cbuf, "%s%s", stk0->cbuf, a);
    }

    if (k > 0) {
	if (_mc)
	    sprintf(stk0->cbuf, "%s,", stk0->cbuf);
	if (ft->tflg[0] & nf_BYREF)
	    sprintf(stk0->cbuf, "%s&", stk0->cbuf);
	sprintf(stk0->cbuf, "%s%s", stk0->cbuf, popb(stk1, &tt));
	if (k > 1)
	    for (i = 0; i < k - 1; i++) {
		sprintf(stk0->cbuf, "%s,", stk0->cbuf);
		if (ft->tflg[i + 1] & nf_BYREF)
		    sprintf(stk0->cbuf, "%s&", stk0->cbuf);
		sprintf(stk0->cbuf, "%s%s", stk0->cbuf, popb(stk1, &tt));
	    }
    }


    if (ft->pceo > k)
	msgrep(mERR, "Too few arguments for %s subroutine", nm);
    if (ft->pc < k)
	msgrep(mERR, "Too many arguments for %s subroutine", nm);

    if (ft->pc > k) {
	if (!k) {
	    sprintf(stk0->cbuf, "%s%s", stk0->cbuf, ft->opt[k - ft->pceo]);
	    k++;
	}
	while (ft->pc > k) {
	    sprintf(stk0->cbuf, "%s,%s", stk0->cbuf,
		    ft->opt[k - ft->pceo]);
	    k++;
	}
    }


    sprintf(stk0->cbuf, "%s)", stk0->cbuf);
    stk0->cbuf += strlen(stk0->cbuf) + 1;
    pushbt(stk0, t);
    return t;
}


etp fune_call(const char *nm, int k, int j)
{
    j = look_overloads(j, k);
    return fune_call_nov(nm, k, j);
}

void exit_sub()
{
    if (cscope != -1)
	msgrep(mERR, "EXIT SUB/FUNCTION outside SUB or FUNCTION");
    fprintf(out, "goto _llf_%d;\n", func);
    lu++;
}

void call_write()
{
    etp t;
    fprintf(out, "%s;\n", popb(stk0, &t));
}

/*
looks for the best match for given function name, arguments (on the stack)
and arguments count. this code is also scope-aware.
it takes argument types directly from the stack, the best possible match
is always found. penalty() returns penalty points for resolving types and type 
casting conflicts.
nice job!
*/

long look_overloads(long j, long p)
{
    int i, pen1 = 0, pen2 = 0, sj;
    long my_bp, f = -1;
    fun_t *ft;
    sj = j;
    for (i = sj + 1; i <= symp; i++)
	if (!strcasecmp(symtab[i].name, symtab[j].name)
	    && (symtab[j].scope == symtab[i].scope)
	    && (symtab[i].ttp.cl == cFUN)) {
	    f = i;
	    ft = (fun_t *) symtab[f].adata;
	    if ((p < ft->pceo) || (p > ft->pc))
		continue;

	    begin_scan(stk1);
	    my_bp = 0;
	    while ((my_bp < p) && (my_bp < ft->pc)) {
		pen2 += penalty(stk1->buf2[my_bp], ft->typ[p - my_bp - 1]);
		my_bp++;
	    }

	    ft = (fun_t *) symtab[j].adata;
	    begin_scan(stk1);
	    my_bp = 0;
	    while ((my_bp < p) && (my_bp < ft->pc)) {
		pen1 += penalty(stk1->buf2[my_bp], ft->typ[p - my_bp - 1]);
		my_bp++;
	    }

	    if (pen2 <= pen1)
		j = f;
	    pen1 = 0;
	    pen2 = 0;
	}
    return j;
}

long look_strict(char *nm, int scope, fun_t * fj)
{
    int i;
    long my_bp, f = -1;
    fun_t *ft;
    for (i = 0; i <= symp; i++)
	if (!strcmp(symtab[i].name, nm)
	    && (scope == symtab[i].scope)
	    && (symtab[i].ttp.cl == cFUN)) {
	    f = i;
	    ft = (fun_t *) symtab[f].adata;
	    if ((fj->pceo != ft->pceo) || (fj->pc != ft->pc))
		continue;

	    my_bp = 0;
	    while (my_bp < ft->pc) {
		if (penalty(ft->typ[my_bp], ft->typ[my_bp]))
		    continue;
		my_bp++;
	    }
	    return i + 1;

	}
    return 0;
}



int penalty(etp a, etp b)
{
    if ((a.cl == b.cl) || (a.cl == cCONST && b.cl == cSCALAR)) {
	if (a.tp == b.tp)
	    return 0;
	if (tIS_INT(a.tp) && tIS_INT(b.tp)) {
	    if (a.tp < b.tp)
		return 1;
	    else
		return 2;
	}
	if (tIS_DBL(a.tp) && tIS_DBL(b.tp)) {
	    if (a.tp < b.tp)
		return 1;
	    else
		return 2;
	}
	if (tIS_INT(a.tp) && tIS_DBL(b.tp))
	    return 2;
	if (tIS_DBL(a.tp) && tIS_INT(b.tp))
	    return 3;
	return 100;
    }
    return 100;
}

long look_sub(char *name, int scope, int p)
{
    long i, f = -1, j = -1;
    fun_t *ft;

    for (i = 0; i <= symp; i++)
	if (!strcasecmp(symtab[i].name, name)
	    && (scope == symtab[i].scope || symtab[i].scope == 1)) {
	    j = i;
	    break;
	};

    if (j == -1)
	return -1;
    ft = (fun_t *) symtab[j].adata;
    if ((p >= ft->pceo) && (p <= ft->pc))
	return j;

    for (i = j + 1; i <= symp; i++)
	if (!strcasecmp(symtab[i].name, name)
	    && (scope == symtab[i].scope)
	    && (symtab[i].ttp.cl == cFUN)) {
	    f = i;
	    ft = (fun_t *) symtab[f].adata;
	    if ((p >= ft->pceo) && (p <= ft->pc))
		return f;
	}
    return -1;
}
